/* Dealing with Unix FTP servers.
   Copyright (C) 1995,96  Hrvoje Niksic
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif
#ifdef TIME_WITH_SYS_TIME
#  include <sys/time.h>
#  include <time.h>
#else /* not TIME_WITH_SYS_TIME_H */
#ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
#else /* not HAVE_SYS_TIME_H */
#  include <time.h>
#endif /* HAVE_SYS_TIME_H */
#endif /* TIME_WITH_SYS_TIME_H */
#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif
#include <ctype.h>
#include <errno.h>
#include <assert.h>

#include "wget.h"
#include "url.h"
#include "utils.h"
#include "ftp.h"
#include "options.h"

extern struct options opt;

/* Converts symbolic permissions to number-style ones, e.g. string
   rwxr-xr-x to 755. For now, it knows nothing of
   setuid/setgid/sticky. ACLs are ignored. */
int
symperms(const char *s)
{
   int perms = 0, i;
   
   if (strlen(s) < 9)
      return 0;
   for (i = 0; i < 3; i++, s += 3)
   {
      perms <<= 3;
      perms += (((*s == 'r') << 2) + ((s[1] == 'w') << 1) +
                (s[2] == 'x' || s[2] == 's'));
   }
   return perms;
}

/* The function parses the Un*x-ish style directory listing stored in
   file, and returns the linked list of fileinfo (system-independent)
   entries. The file is considered to be produced by the standard Unix
   ls -la output (whatever that meant:). BSD (no group) and SYSV (with
   group) listings are handled.

   The time stamps are stored in a separate variable, time_t
   compatible (I hope). The time-zones are ignored (unfortunately). */
struct fileinfo *
ftp_parse_unix_ls(const char *file)
{
   FILE *fp;
   static const char *months[] = {
      "Jan", "Feb", "Mar", "Apr", "May", "Jun",
      "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
   };
   int next, len, fnlen, i, error, ignore;
   int year, month, day;        /* For time analysis.  */
   int hour, min;
   struct tm timestruct, *tnow;
   time_t timenow;
   
   char *line, *tok;            /* Tokenizer. */
   struct fileinfo *dir, *l, cur; /* List creation. */

   fp = fopen(file, "r");
   if (!fp)
   {
      fprintf(opt.lfile, "%s: %s\n", file, mystrerror(errno));
      return NULL;
   }
   dir = l = NULL;

   /* Line loop to end of file: */
   while ((line = read_whole_line(fp)))
   {
#ifdef DEBUG
      if (opt.debug)
         fprintf(opt.lfile, "%s\n", line);
#endif
      len = strlen(line);
      /* Destroy <CR> if there is one. */
      if (len && line[len - 1] == '\r')
         line[len - 1] = '\0';

      /* Skip if total... */
      if (!strncmp(line, "total", 5))
      {
         free(line);
         continue;
      }
      /* Get the first token (permissions). */
      tok = strtok(line, " ");
      if (!tok)
      {
         free(line);
         continue;
      }

      cur.name = NULL;
      cur.linkto = NULL;

      /* Decide whether we deal with a file or a directory. */
      switch (*tok)
      {
         case '-':
            cur.type = PLAINFILE;
            DEBUGP("PLAINFILE; ");
            break;
         case 'd':
            cur.type = DIRECTORY;
            DEBUGP("DIRECTORY; ");
            break;
         case 'l':
            cur.type = SYMLINK;
            DEBUGP("SYMLINK; ");
            break;
         default:
            cur.type = UNKNOWN;
            DEBUGP("UNKOWN; ");
            break;
      }

      cur.perms = symperms(tok + 1);
#ifdef DEBUG
      if (opt.debug)
         fprintf(opt.lfile, "perms %0o; ", cur.perms);
#endif
      
      error = ignore = 0;       /* Errnoeous and ignoring entries are
                                   treated equally for now. */
      year = hour = min = 0;    /* Silence the compiler. */
      month = day = 0;
      next = -1;
      /* While there are tokens on the line, parse them. Next is the
         number of tokens left until the filename. The "safe" place
         was decided to be the month name. When a month name is
         encountered, next is set to 5. Also, the preceding characters
         are parsed to get the file size. */
      while ((tok = strtok(NULL, " ")))
      {
         --next;
         if (next < 0)          /* I.e. a month name was not
                                   encountered. */
         {
            for (i = 0; i < 12; i++)
               if (!strcmp(tok, months[i]))
                  break;
            /* If we got a month, it means the token before it is the
               size, and the filename is three tokens away. */
            if (i != 12)
            {
               char *t = tok - 2;
               long mul = 1;
               
               for (cur.size = 0; t > line && isdigit(*t); mul *= 10, t--)
                  cur.size += mul * (*t - '0');
               if (t == line)
               {
                  /* Something is seriously wrong. */
                  error = 1;
                  break;
               }
               month = i;
               next = 5;
#ifdef DEBUG
               if (opt.debug)
                  fprintf(opt.lfile, "month: %s; ", months[month]);
#endif
            }
         }
         if (next == 4)         /* Days */
         {
            if (tok[1])         /* Two-digit... */
               day = 10 * (*tok - '0') + tok[1] - '0';
            else                /* ...or one-digit. */
               day = *tok - '0';
#ifdef DEBUG
            if (opt.debug)
               fprintf(opt.lfile, "day: %d; ", day);
#endif
         }
         else if (next == 3)
         {
            if (tok[2] == ':')  /* hh:mm */
            {
               min = (tok[4] - '0') + 10 * (tok[3] - '0');
               hour = (tok[1] - '0') + 10 * (*tok - '0');
               year = 0;
            }
            else if (tok[1] == ':')  /* h:mm */
            {
               min = (tok[3] - '0') + 10 * (tok[2] - '0');
               hour = (*tok - '0');
               year = 0;
            }
            else                /* yyyy or yy */
            {
               year = atoi(tok);
               min = hour = 0;
            }
#ifdef DEBUG
            if (opt.debug)
            {
               if (year)
                  fprintf(opt.lfile, "year: %d (no tm); ", year);
               else
                  fprintf(opt.lfile, "time: %d:%d (no yr); ", hour, min);
            }
#endif
         }
         else if (next == 2)    /* The file name */
         {
            /* If we have the filename, add it to the list of files or
               directories. */
            /* "." and ".." are an exception! */
            if (!strcmp(tok, ".") || !strcmp(tok, ".."))
            {
               DEBUGP("\nIgnoring `.' and `..'; ");
               ignore = 1;
               break;
            }
            /* Some FTP sites choose to have ls -F as their default
               LIST output, which marks the symlinks with a trailing
               '@', directory names with a trailing '/' and
               executables with a trailing '*'. This is no problem
               unless encountering a symbolic link ending with '@', or
               an executable ending with '*' on a server *without*
               default -F output.  I believe these cases are very
               rare.  */
            fnlen = strlen(tok);    /* We'll need the length, so no need
                                       to strdup(). */
            cur.name = nmalloc(fnlen + 1);
            memcpy(cur.name, tok, fnlen + 1);
            if (fnlen)
            {
               if (cur.type == DIRECTORY && cur.name[fnlen - 1] == '/')
               {
                  cur.name[fnlen - 1] = '\0';
                  DEBUGP("trailing `/' on dir.\n");
               }
               else if (cur.type == SYMLINK && cur.name[fnlen - 1] == '@')
               {
                  cur.name[fnlen - 1] = '\0';
                  DEBUGP("trailing `@' on link.\n");
               }
               else if (cur.type == PLAINFILE
                        && (cur.perms & 0111)
                        && cur.name[fnlen - 1] == '*')
               {
                  cur.name[fnlen - 1] = '\0';
                  DEBUGP("trailing `*' on exec.\n");
               }
            }
            if (cur.type != SYMLINK)
               break;
         }
         else if (next == 1)    /* Symlink name. */
         {
            if (strcmp(tok, "->") != 0)
               break;
         }
         else if (!next)
         {
            /* The following implementation imposes a limit that there
               may be no spaces in a symlink pointer. This may be
               fixed in the future. */
            cur.linkto = nstrdup(tok);
#ifdef DEBUG
            if (opt.debug)
               fprintf(opt.lfile, "link to: %s\n", cur.linkto);
#endif
         }
      }

      if (!cur.name || (cur.type == SYMLINK && !cur.linkto))
         error = 1;
      
      DEBUGP("\n");

      if (error || ignore)
      {
         DEBUGP("Skipping.\n");
         if (cur.name)
            free(cur.name);
         if (cur.linkto)
            free(cur.linkto);
         free(line);
         continue;
      }
      
      if (!dir)
      {
         l = dir = (struct fileinfo *)nmalloc(sizeof(struct fileinfo));
         memcpy(l, &cur, sizeof(cur));
         l->prev = l->next = NULL;
      }
      else
      {
         cur.prev = l;
         l->next = (struct fileinfo *)nmalloc(sizeof(struct fileinfo));
         l = l->next;
         memcpy(l, &cur, sizeof(cur));
         l->next = NULL;
      }
      /* Get the current time. */
      timenow = time(NULL);
      tnow = localtime(&timenow);
      /* Build the time-stamp (adapted from something by
         zaga@fly.cc.fer.hr). */
      timestruct.tm_sec   = 0;
      timestruct.tm_min   = min;
      timestruct.tm_hour  = hour;
      timestruct.tm_mday  = day;
      timestruct.tm_mon   = month;
      if (year == 0)
         timestruct.tm_year = tnow->tm_year;
      else
         timestruct.tm_year = year;
      if (timestruct.tm_year >= 1900)
         timestruct.tm_year -= 1900;
      timestruct.tm_wday  = 0;
      timestruct.tm_yday  = 0;
      timestruct.tm_isdst = -1;
      l->tstamp = mktime(&timestruct); /* Store the time-stamp. */
      
      /* Free the line. */
      free(line);
   }
   /* Close the file. */
   fclose(fp);
   return dir;
}

