/* Reading/parsing the initialization file.
   Copyright (C) 1995,96  Hrvoje Niksic
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif
#ifdef HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif
#include <pwd.h>

#include "wget.h"
#include "options.h"
#include "utils.h"
#include "init.h"
#include "url.h"
#include "ftp.h"
#include "host.h"
#include "recur.h"

extern char *version_string;
extern struct options opt;
extern int errno;

/* List of recognized commands, each consisting of a string-name and
   ID. */
struct cmd commands[] = {
   { "accept", ACCEPT },
   { "addhostdir", ADDHOSTDIR },
   { "alwaysrest", ALWAYSREST },
   { "backups", BACKUPS },
   { "basehref", BASEHREF },
   { "cache", CACHE },
   { "convertlinks", CONVERTLINKS },
   { "debug", DEBUG_ },
   { "dirmode", DIRMODE },
   { "dirprefix", DIRPREFIX },
   { "dirstruct", DIRSTRUCT },
   { "domains", DOMAINS },
   { "dotbytes", DOTBYTES },
   { "dotsinline", DOTSINLINE },
   { "dotspacing", DOTSPACING },
   { "dotstyle", DOTSTYLE },
   { "excludedirectories", EXCLUDEDIRECTORIES },
   { "fileprefix", FILEPREFIX },
   { "followftp", FOLLOWFTP },
   { "forcehtml", FORCEHTML },
   { "ftpproxy", FTPPROXY },
   { "glob", GLOB },
   { "header", HEADER },
   { "htmlify", HTMLIFY },
   { "httppasswd", HTTPPASSWD },
   { "httpproxy", HTTPPROXY },
   { "httpuser", HTTPUSER },
   { "includedirectories", INCLUDEDIRECTORIES },
   { "input", INPUT },
   { "killlonger", KILLLONGER },
   { "logfile", LOGFILE },
   { "login", LOGIN },
   { "mirror", MIRROR },
   { "noclobber", NOCLOBBER },
   { "noparent", NOPARENT },
   { "noproxy", NOPROXY },
   { "numtries", NUMTRIES },
   { "outputdocument", OUTPUTDOCUMENT },
   { "passwd", PASSWD },
   { "prefixfiles", PREFIXFILES },
   { "quiet", QUIET },
   { "quota", QUOTA },
   { "reclevel", RECLEVEL },
   { "recursive", RECURSIVE },
   { "reject", REJECT },
   { "relativeonly", RELATIVEONLY },
   { "removelisting", REMOVELISTING },
   { "robots", ROBOTS },
   { "saveheaders", SAVEHEADERS },
   { "serverresponse", SERVERRESPONSE },
   { "simplehostcheck", SIMPLEHOSTCHECK },
   { "spanhosts", SPANHOSTS },
   { "spider", SPIDER },
   { "timeout", TIMEOUT },
   { "timestamping", TIMESTAMPING },
   { "useproxy", USEPROXY },
   { "useragent", USERAGENT },
   { "verbose", VERBOSE },
   { NULL, LAST }
};


/* Return command ID of a valid command, -1 if none. */
cmdid
comind(const struct cmd *commands, const char *com)
{
   int i;

   for (i = 0; commands[i].name; i++)
      if (!strcasecmp(commands[i].name, com))
         return commands[i].id;
   return NONE;
}

/* Reset the variables to default values. */
void
defaults(void)
{
   char *tmp;
   
#ifdef DEBUG
   opt.debug = 0;
#endif
   opt.verbose = -1;
   opt.quiet = 0;
   
   opt.kill_longer = 0;
   opt.lfilename = NULL;
   opt.input_filename = NULL;
   opt.prefix_files = 0;
   opt.dirstruct = 0;
   opt.no_dirstruct = 0;
   opt.fileprefix = nstrdup(DEFAULT_FILE_PREFIX);
   opt.dir_prefix = nstrdup(DEFAULT_DIRPREFIX);
   
   opt.ntry = DEFAULT_NUMBER_OF_TRIES;
   opt.recursive = 0;
   opt.maxreclevel = DEFAULT_MAXRECLEVEL;
   opt.spanhost = 0;
   opt.relative_only = 0;
   opt.no_parent = 0;
   opt.accepts = opt.rejects = NULL;
   opt.includes = opt.excludes = NULL;
   
   opt.simple_check = 0;
   opt.dirmode = DEFAULT_DIRMODE;
   opt.add_hostdir = 1;
   opt.noclobber = 0;
   
   opt.follow_ftp = 0;
   opt.always_rest = 0;
   opt.ftp_acc = nstrdup(DEFAULT_FTP_ACCT);
   opt.ftp_pass = nstrdup(ftp_getaddress());
   opt.ftp_glob = 1;
   opt.htmlify = 1;

   opt.http_user = NULL;
   opt.http_passwd = NULL;
   opt.user_header = NULL;

   opt.useragent = NULL;
   opt.output_document = NULL;

   opt.use_proxy = 1;
   opt.http_proxy = opt.ftp_proxy = NULL;
   opt.no_proxy = NULL;
   tmp = getenv("no_proxy");
   if (tmp)
      opt.no_proxy = sepstring(tmp);
   opt.proxy_cache = 1;
   opt.base_href = NULL;

#ifdef HAVE_SELECT
   opt.timeout = DEFAULT_TIMEOUT;
#endif
   
   opt.use_robots = 1;

   opt.quota = 0;
   opt.downloaded = 0;
   opt.numurls = 0;

   opt.print_server_response = 0;
   opt.save_headers = 0;
   opt.timestamping = 0;
   opt.remove_listing = 1;

   opt.backups = 0;
   opt.convert_links = 0;

   opt.dot_bytes = DOT_BYTES;
   opt.dot_spacing = DOT_SPACING;
   opt.dots_in_line = DOTS_IN_LINE;
}

/* Return the path to the user's .wgetrc. */
char *
init_path(void)
{
   char *env, *home, *path;
   struct stat buf;
   int err;
   struct passwd *pwd;

   /* Try the environment. */
   if ((env = getenv("WGETRC")))
   {
      path = nstrdup(env);
      err = stat(path, &buf);
      if (err == 0)
         return path;
      free(path);
      fprintf(stderr, "Cannot load .wgetrc.\n");
      exit(1);
   }
   /* If that failed, try $HOME/.wgetrc. */
   home = getenv("HOME");
   /* If HOME is not defined, try getting it from the password file. */
   if (!home)
   {
      pwd = getpwuid(getuid());
      if (!pwd || !pwd->pw_dir)
         return NULL;
      home = pwd->pw_dir;
   }
   if (!home)
      return NULL;
   path = (char *)nmalloc(strlen(home) + 1 + strlen(".wgetrc") + 1);
   sprintf(path, "%s/.wgetrc", home);
   err = stat(path, &buf);
   if (err == -1)
   {
      free(path);
      return NULL;
   }
   return path;
}

/* Initialize variables from a wgetrc file */
void
run_wgetrc(const char *path)
{
   char *com, *val;
   FILE *fp;
   char *line;
   int stat, ln, length;

   fp = fopen(path, "r");
   if (!fp)
   {
      fprintf(stderr, "Cannot read %s (%s).\n", path, mystrerror(errno));
      return;
   }
   /* Reset line number. */
   ln = 1;
   /* While there are lines in the file: */
   while ((line = read_whole_line(fp)))
   {
      length = strlen(line);
      if (length && line[length - 1] == '\r')
         line[length - 1] = '\0';
      /* Parse the line. */
      stat = parse_line(line, &com, &val);
      free(line);
      /* If everything is OK, set the value. */
      if (stat == 1)
      {
         if (!setval(com, val))
            fprintf(stderr, "Syntax error in line %d.\n", ln);
         free(com);
         free(val);
      }
      else if (stat == 0)
         fprintf(stderr, "Syntax error in line %d.\n", ln);
      ++ln;
   }
   fclose(fp);
}

/* Initialize the defaults and run the system wgetrc and user's own
   wgetrc. */
void
initialize(void)
{
   char *path;
   struct stat buf;
   int err;

   /* Load defaults. */
   defaults();

   /* If SYSTEM_WGETRC is defined, use it. */
#ifdef SYSTEM_WGETRC
   err = stat(SYSTEM_WGETRC, &buf);
   if (err != -1) /* File exists */
      run_wgetrc(SYSTEM_WGETRC);
#endif
   /* Override it with your own, if one exists. */
   if (!(path = init_path()))
      return;
   run_wgetrc(path);
   free(path);
}

/* Parse the line pointed by line, with the syntax:
   <sp>* command <sp>* = <sp>* value <newline>
   Uses malloc to allocate space for command and value.
   If the line is invalid, data is freed and 0 is returned.

   Return values:
    1 - success
    0 - failure
   -1 - comment */
int
parse_line(const char *line, char **com, char **val)
{
   int i, j;

   /* Skip spaces. */
   for (i = 0; line[i] == ' ' || line[i] == '\t'; i++);

   /* Skip empty or hashed lines. */
   if (!line[i] || line[i] == '\n' || line[i] == '#')
      return -1;
   
   /* Allocate command. */
   *com = (char *)nmalloc(1);
   (*com)[0] = '\0';
   for (j = 0; (isalpha(line[i]) || line[i] == '_' || line[i] == '-'); i++)
   {
      if (line[i] == '_' || line[i] == '-')
         continue;
      *com = (char *)nrealloc(*com, j + 2);
      (*com)[j] = tolower(line[i]);
      (*com)[j + 1] = '\0';
      ++j;
   }
   /* The next char should be space or '='. */
   if (!isspace(line[i]) && (line[i] != '='))
   {
      free(*com);
      return 0;
   }
   /* If the command is invalid, exit now. */
   if (comind(commands, *com) == NONE)
   {
      free(*com);
      return 0;
   }
   /* Skip spaces before '='. */
   for (; isspace(line[i]); i++);
   /* If '=' not found, bail out. */
   if (line[i] != '=')
   {
      free(com);
      return 0;
   }
   /* Skip spaces after '='. */
   for (++i; isspace(line[i]); i++);
   /* Get the ending position. */
   for (j = i; line[i] && line[i] != '\n'; i++);
   /* Allocate *val, and copy from line. */
   (*val) = (char *)nmalloc(i - j + 1);
   strncpy(*val, line + j, i - j);
   (*val)[i - j] = '\0';
   return 1;
}

/* Do the job for each line.  No fatals -- error signal prints a
   warning and resets to default value.  All error messages are
   printed to stderr, *not* to opt.lfile, since opt.lfile wasn't even
   generated yet.  */
int
setval(const char *com, const char *val)
{
   int flag, mirror;
   int i;                       /* To be used for various loops. */
   const char *p;               /* Likewise. */
   cmdid cmd;

   if (!com || !val)
      return 0;
   cmd = comind(commands, com);
   if (cmd == NONE)
   {
#ifdef DEBUG
      fprintf(stderr, "Wrong var %s, value %s.\n", com, val);
#endif
      return 0;
   }
   switch (cmd)
   {
      case ACCEPT:
         if (*val)
            opt.accepts = merge_vecs(opt.accepts, sepstring(val));
         else
         {
            free_vec(opt.accepts);
            opt.accepts = NULL;
         }
         break;
      case ADDHOSTDIR:
         return setonoff(&opt.add_hostdir, com, val);
         break;
      case ALWAYSREST:
         return setonoff(&opt.always_rest, com, val);
         break;
      case BACKUPS:
         return setonoff(&opt.backups, com, val);
         break;
      case CACHE:
         return setonoff(&opt.proxy_cache, com, val);
         break;
      case BASEHREF:
         if (opt.base_href)
            free(opt.base_href);
         opt.base_href = nstrdup(val);
         break;
      case CONVERTLINKS:
         return setonoff(&opt.convert_links, com, val);
         break;
      case DEBUG_:
#ifdef DEBUG
         return setonoff(&opt.debug, com, val);
#else
         fprintf(stderr, "%s: debug support not loaded.\n", com);
         return 0;
#endif
         break;
      case DIRMODE:
         flag = getperms(val);
         if (flag == -1)
         {
            fprintf(stderr, "%s: Invalid mode specification `%s'.\n",
                    com, val);
            return 0;
         }
         opt.dirmode = DEFAULT_DIRMODE;
         break;
      case DIRPREFIX:
         free(opt.dir_prefix);
         opt.dir_prefix = nstrdup(val);
         break;
      case DIRSTRUCT:
         if (!setonoff(&opt.dirstruct, com, val))
            return 0;
         /* Since dirstruct behaviour is explicitly changed,
            no_dirstruct must be affected inversely. */
         if (opt.dirstruct)
            opt.no_dirstruct = 0;
         else
            opt.no_dirstruct = 1;
         break;
      case DOMAINS:
         if (*val)
            opt.domains = merge_vecs(opt.domains, sepstring(val));
         else
         {
            free_vec(opt.domains);
            opt.accepts = NULL;
         }
         break;
      case DOTBYTES:
         return setbytes(&opt.dot_bytes, com, val);
         break;
      case DOTSINLINE:
         return setnuminf(&opt.dots_in_line, com, val);
         break;
      case DOTSPACING:
         return setnuminf(&opt.dot_spacing, com, val);
         break;
      case DOTSTYLE:
         /* Retrieval styles. */
         if (!strcasecmp(val, "default"))
         {
            /* Default style: 1K dots, 10 dots in a cluster, 50 dots
               in a line. */
            opt.dot_bytes = DOT_BYTES;
            opt.dot_spacing = DOT_SPACING;
            opt.dots_in_line = DOTS_IN_LINE;
         }
         else if (!strcasecmp(val, "binary"))
         {
            /* "Binary" retrieval: 8K dots, 16 dots in a cluster, 64
               dots (512K) in a line. */
            opt.dot_bytes = 8192;
            opt.dot_spacing = 16;
            opt.dots_in_line = 64;
         }
         else if (!strcasecmp(val, "mega"))
         {
            /* "Mega" retrieval, for retrieving very long files; each
               dot is 64K, 8 dots in a cluster, 6 clusters (3M) in a
               line.  */
            opt.dot_bytes = 65536L;
            opt.dot_spacing = 8;
            opt.dots_in_line = 48;
         }
         else if (!strcasecmp(val, "micro"))
         {
            /* "Micro" retrieval, for retrieving very small files;
               each dot is 128 bytes, 8 dots in a cluster, 6 clusters
               (6K) in a line.  */
            opt.dot_bytes = 128;
            opt.dot_spacing = 8;
            opt.dots_in_line = 48;
         }
         else
         {
            fprintf(stderr, "%s: Invalid specification `%s'.\n", com, val);
            return 0;
         }
         break;
      case EXCLUDEDIRECTORIES:
         if (*val)
            opt.excludes = merge_vecs(opt.excludes, sepstring(val));
         else
         {
            free_vec(opt.excludes);
            opt.excludes = NULL;
         }
         break;
      case FILEPREFIX:
         free(opt.fileprefix);
         opt.fileprefix = nstrdup(val);
         break;
      case FOLLOWFTP:
         return setonoff(&opt.follow_ftp, com, val);
         break;
      case FORCEHTML:
         return setonoff(&opt.force_html, com, val);
         break;
      case FTPPROXY:
         if (opt.ftp_proxy)
            free(opt.ftp_proxy);
         opt.ftp_proxy = nstrdup(val);
         break;
      case GLOB:
         return setonoff(&opt.ftp_glob, com, val);
         break;
      case HEADER:
         if (!*val)
         {
            /* Empty header means to reset headers. */
            if (opt.user_header)
            {
               free(opt.user_header);
               opt.user_header = NULL;
            }
         }
         else
         {
            /* Check for sanity of the header value. */
            flag = 0;
            for (p = val; *p && *p != ':' && !isspace(*p); p++);
            /* The header MUST contain ':' preceded by at least one
               non-space character. */
            if (*p != ':' || p == val)
               flag = 1;
            /* The header MUST NOT contain newlines. */
            if (!flag)
               if (strchr(val, '\n'))
                  flag = 1;
            if (flag)
            {
               fprintf(stderr, "%s: Invalid specification `%s'.\n", com, val);
               return 0;
            }
            i = opt.user_header ? strlen(opt.user_header) : 0;
            opt.user_header = (char *)nrealloc(opt.user_header,
                                               i + strlen(val) + 2 + 1);
            strcpy(opt.user_header + i, val);
            i += strlen(val);
            opt.user_header[i++] = '\r';
            opt.user_header[i++] = '\n';
            opt.user_header[i] = '\0';
         }
         break;
      case HTMLIFY:
         flag = setonoff(&opt.htmlify, com, val);
         if (flag && !opt.htmlify)
            opt.remove_listing = 0;
         return flag;
         break;
      case HTTPPASSWD:
         if (opt.http_passwd)
            free(opt.http_passwd);
         opt.http_passwd = nstrdup(val);
         break;
      case HTTPPROXY:
         if (opt.http_proxy)
            free(opt.http_proxy);
         opt.http_proxy = nstrdup(val);
         break;
      case HTTPUSER:
         if (opt.http_user)
            free(opt.http_user);
         opt.http_user = nstrdup(val);
         break;
      case INCLUDEDIRECTORIES:
         if (*val)
            opt.includes = merge_vecs(opt.includes, sepstring(val));
         else
         {
            free_vec(opt.includes);
            opt.includes = NULL;
         }
         break;
      case INPUT:
         opt.input_filename = nstrdup(val);
         break;
      case KILLLONGER:
         return setonoff(&opt.kill_longer, com, val);
         break;
      case LOGFILE:
         if (opt.lfilename)
            free(opt.lfilename);
         opt.lfilename = nstrdup(val);
         break;
      case LOGIN:
         free(opt.ftp_acc);
         opt.ftp_acc = nstrdup(val);
         break;
      case MIRROR:
         if (!setonoff(&mirror, com, val))
            return 0;
         if (mirror)
         {
#ifndef NO_RECURSION
            opt.recursive = 1;
            if (!opt.no_dirstruct)
               opt.dirstruct = 1;
            opt.timestamping = 1;
            opt.maxreclevel = 0;
            opt.remove_listing = 0;
#else
            fprintf(stderr, "Recursion is not enabled.\n");
#endif
         }
         break;
      case NOCLOBBER:
         return setonoff(&opt.noclobber, com, val);
         break;
      case NOPARENT:
         return setonoff(&opt.no_parent, com, val);
         break;
      case NOPROXY:
         if (*val)
            opt.no_proxy = merge_vecs(opt.no_proxy, sepstring(val));
         else
         {
            free_vec(opt.no_proxy);
            opt.no_proxy = NULL;
         }
         break;
      case NUMTRIES:
         return setnuminf(&opt.ntry, com, val);
         break;
      case OUTPUTDOCUMENT:
         if (opt.output_document)
            free(opt.output_document);
         opt.output_document = nstrdup(val);
         opt.ntry = 1;
         if (ISHYPHEN(opt.output_document))
            opt.quiet = 1;
         break;
      case PASSWD:
         free(opt.ftp_pass);
         opt.ftp_pass = nstrdup(val);
         break;
      case PREFIXFILES:
         return setonoff(&opt.prefix_files, com, val);
         break;
      case QUIET:
         return setonoff(&opt.quiet, com, val);
         break;
      case QUOTA:
         return setbytes(&opt.quota, com, val);
         break;
      case RECLEVEL:
         return setnuminf(&opt.maxreclevel, com, val);
         break;
      case RECURSIVE:
#ifndef NO_RECURSION
         if (!setonoff(&opt.recursive, com, val))
            return 0;
         else
         {
            if (opt.recursive && !opt.no_dirstruct)
               opt.dirstruct = 1;
         }
         break;
#else
         fprintf(stderr, "Recursion is not enabled.\n");
#endif
         break;
      case REJECT:
         if (*val)
            opt.rejects = merge_vecs(opt.rejects, sepstring(val));
         else
         {
            free_vec(opt.rejects);
            opt.rejects = NULL;
         }
         break;
      case RELATIVEONLY:
         return setonoff(&opt.relative_only, com, val);
         break;
      case REMOVELISTING:
         return setonoff(&opt.remove_listing, com, val);
         break;
      case ROBOTS:
         return setonoff(&opt.use_robots, com, val);
         break;
      case SAVEHEADERS:
         return setonoff(&opt.save_headers, com, val);
         break;
      case SERVERRESPONSE:
         return setonoff(&opt.print_server_response, com, val);
         break;
      case SIMPLEHOSTCHECK:
         return setonoff(&opt.simple_check, com, val);
         break;
      case SPANHOSTS:
         return setonoff(&opt.spanhost, com, val);
         break;
      case SPIDER:
         return setonoff(&opt.spider, com, val);
         break;
      case TIMEOUT:
#ifdef HAVE_SELECT
         return setnuminf(&opt.timeout, com, val);
#else
         fprintf(stderr, "Unable to set timeout without select().\n");
#endif
         break;
      case TIMESTAMPING:
         return setonoff(&opt.timestamping, com, val);
         break;
      case USERAGENT:
         if (opt.useragent)
            free(opt.useragent);
         opt.useragent = nstrdup(val);
         break;
      case USEPROXY:
         return setonoff(&opt.use_proxy, com, val);
         break;
      case VERBOSE:
         return setonoff(&opt.verbose, com, val);
         break;
      case NONE: case LAST: default:
         fprintf(stderr, "Error condition, please investigate.\n");
         exit(1);
   }
   return 1;
}

/* Return 0 if off, 1 if on, -1 if error. */
int
onoff(const char *val)
{
   if (strcasecmp(val, "on") == 0 || strcmp(val, "1") == 0)
      return 1;
   else if (strcasecmp(val, "off") == 0 || strcmp(val, "0") == 0)
      return 0;
   else
      return -1;
}

/* Set the integer value pointed to by what on value represented in
   string val. String com is used only for error messages. */
int
setonoff(int *what, const char *com, const char *val)
{
   int flag;

   flag = onoff(val);
   if (flag == -1)
   {
      fprintf(stderr, "%s: Please specify on or off.\n", com);
      return 0;
   }
   *what = flag;
   return 1;
}

/* Similar to setonoff, but with "inf" meaning zero.  If the
   specification is incorrect, the number remains unchanged.  */
int
setnuminf(int *what, const char *com, const char *val)
{
   int res;
   
   if (!strcasecmp(val, "inf"))
   {
      *what = 0;
      return 1;
   }
   res = myatoi(val);
   if (res == -1)
   {
      fprintf(stderr, "%s: Invalid specification `%s'.\n", com, val);
      return 0;
   }
   *what = res;
   return 1;
}

/* Return the int value of a positive integer written in a string,
   or -1 if an error was encountered */
int
myatoi(const char *s)
{
   int res;
   const char *orig;

   orig = s;
   for (res = 0; *s && isdigit(*s); s++)
   {
      res *= 10;
      res += *s - '0';
   }
   if (*s || orig == s)
      return -1;
   else
      return res;
}

/* Return the decimal representation of octal mode three-digit
   permission string, or -1 if incorrect. */
int
getperms(const char *s)
{
   if (strlen(s) != 3 || !ISODIGIT(s[0]) || !ISODIGIT(s[1]) || !ISODIGIT(s[2]))
      return -1;
   return (s[2] - '0') + 8 * (s[1] - '0') + 64 * (s[0] - '0');
}

/* Free the memory allocated by global variables. */
void
cleanup(void)
{
   recursive_retrieve(NULL, NULL, RCLEANUP);
   clean_hosts();
   if (opt.dfp)
      fclose(opt.dfp);
   if (opt.lfilename)
      free(opt.lfilename);
   free(opt.dir_prefix);
   free(opt.fileprefix);
   if (opt.input_filename)
      free(opt.input_filename);
   if (opt.output_document)
      free(opt.output_document);
   free_vec(opt.accepts);
   free_vec(opt.rejects);
   free_vec(opt.excludes);
   free_vec(opt.includes);
   free_vec(opt.domains);
   free(opt.ftp_acc);
   free(opt.ftp_pass);
   if (opt.ftp_proxy)
      free(opt.ftp_proxy);
   if (opt.http_proxy)
      free(opt.http_proxy);
   free_vec(opt.no_proxy);
   if (opt.useragent)
      free(opt.useragent);
   if (opt.http_user)
      free(opt.http_user);
   if (opt.http_passwd)
      free(opt.http_passwd);
   if (opt.user_header)
      free(opt.user_header);
}

/* Sets the value stored in string val to a long int out, allowing
   several postfices, with the following syntax (regexp):

   [0-9]+       -> bytes
   [0-9]+[kK]   -> bytes * 1024
   [0-9]+[mM]   -> bytes * 1024 * 1024
   inf          -> 0

   Anything else is flagged as incorrect, and *out is unchanged. */
int
setbytes(long *out, const char *com, const char *val)
{
   long result;
   const char *p;

   result = 0;
   p = val;
   /* Check for "inf". */
   if (p[0] == 'i' && p[1] == 'n' && p[2] == 'f' && p[3] == '\0')
   {
      *out = 0L;
      return 1;
   }
   /* Search for digits and construct result. */
   for (; *p && isdigit(*p); p++)
   {
      result *= 10;
      result += *p - '0';
   }
   /* If no digits were found, or more than one character is following
      them, bail out. */
   if (p == val || (*p != '\0' && *(p + 1) != '\0'))
   {
      printf("%s: Invalid specification `%s'\n", com, val);
      return 0;
   }
   /* Search for a designator. */
   switch(*p)
   {
      case 'K': case 'k':
         /* Kilobytes */
         result *= 1024;
         break;
      case 'M': case 'm':
         /* Megabytes */
         result *= (long)1024 * 1024;
         break;
      case '\0':
         /* Never mind. */
         break;
      default:
         printf("%s: Invalid specification `%s'\n", com, val);
         return 0;
   }
   *out = result;
   return 1;
}
